import json
import pickle
from collections.abc import Callable
from enum import IntEnum, auto
from typing import Any

import pytest
from pydantic import BaseModel, Field
from pytest_subtests import SubTests

from aviary.env import DummyEnv
from aviary.tools import (
    INVALID_TOOL_NAME,
    FunctionInfo,
    Tool,
    ToolCall,
    ToolRequestMessage,
    argref_by_name,
)


def simple() -> None:
    """Doing nothing may be better than doing something."""


def intuitive_arg(x: str) -> float:  # type: ignore[empty-body]
    """Cast the input argument x to a float."""


class StubState(BaseModel):
    """Stub model docstring."""

    defaulted_int: int = Field(default=1, description="A description of the int.")
    required_str: str = Field(description="A description of the str.")


class StubEnum(IntEnum):
    """Stub enum docstring."""

    STUB1 = auto()
    STUB2 = auto()


def many_edge_cases(
    x: int,
    y: None,
    union: int | None,
    pydantic_model: StubState,
    basic_dict: dict[str, int],
    complex_dict: dict[str, tuple[str, int]],
    enum: StubEnum,
    defaulted_str: str = "default",
    defaulted_float: float = 1.0,
) -> None:
    """
    Check using docstrings as partial f-string templates like so: {summary_format}.

    Args:
        x: Yes, I end with a colon :
        y: I am null.
            And despite that there is a multiline argument description.
        union: I am a union and the current year is {current_year}.
        pydantic_model: I am a Pydantic model.
        basic_dict: I am a dictionary with primitive values.
        complex_dict: I am a dictionary with complex values.
        enum: I am an enum.
        defaulted_str: I have a string default value.
        defaulted_float: I have a float default value.
    """


def add(a: int, b: int) -> int:
    """Add two numbers.

    Args:
        a: first number
        b: second number

    Returns:
        I am not yet included, perhaps someday I should be.
    """
    return a + b


def example_fxn(x: int, y: str, z: float) -> None:
    r"""A test function.

    There may be non-summary content.

    \f

    I should be ignored.

    Args:
        x: x
        y: y
        z: z
    """
    assert isinstance(x, int)
    assert isinstance(y, str)
    assert isinstance(z, float)


class TestTool:
    @pytest.mark.parametrize(
        ("fn", "kwargs", "expected"),
        [
            pytest.param(
                simple,
                {},
                {
                    "type": "function",
                    "info": {
                        "name": "simple",
                        "description": (
                            "Doing nothing may be better than doing something."
                        ),
                        "parameters": {
                            "properties": {},
                            "required": [],
                            "type": "object",
                        },
                    },
                },
                id="only-summary",
            ),
            pytest.param(
                intuitive_arg,
                {"allow_empty_param_descriptions": True},
                {
                    "type": "function",
                    "info": {
                        "name": "intuitive_arg",
                        "description": "Cast the input argument x to a float.",
                        "parameters": {
                            "properties": {"x": {"title": "X", "type": "string"}},
                            "required": ["x"],
                            "type": "object",
                        },
                    },
                },
                id="only-summary",
            ),
            pytest.param(
                many_edge_cases,
                {"current_year": 2024},  # Intentionally left format_1 unformatted,
                {
                    "type": "function",
                    "info": {
                        "name": "many_edge_cases",
                        "description": (
                            "Check using docstrings as partial f-string templates like"
                            " so: {summary_format}."
                        ),
                        "parameters": {
                            "type": "object",
                            "properties": {
                                "x": {
                                    "description": "Yes, I end with a colon :",
                                    "title": "X",
                                    "type": "integer",
                                },
                                "y": {
                                    "description": (
                                        "I am null. And despite that there is a"
                                        " multiline argument description."
                                    ),
                                    "title": "Y",
                                    "type": "null",
                                },
                                "union": {
                                    "anyOf": [{"type": "integer"}, {"type": "null"}],
                                    "description": (
                                        "I am a union and the current year is 2024."
                                    ),
                                    "title": "Union",
                                },
                                "pydantic_model": {
                                    "allOf": [{"$ref": "#/$defs/StubState"}],
                                    "description": "I am a Pydantic model.",
                                },
                                "basic_dict": {
                                    "additionalProperties": {"type": "integer"},
                                    "description": (
                                        "I am a dictionary with primitive values."
                                    ),
                                    "title": "Basic Dict",
                                    "type": "object",
                                },
                                "complex_dict": {
                                    "additionalProperties": {
                                        "maxItems": 2,
                                        "minItems": 2,
                                        "prefixItems": [
                                            {"type": "string"},
                                            {"type": "integer"},
                                        ],
                                        "type": "array",
                                    },
                                    "description": (
                                        "I am a dictionary with complex values."
                                    ),
                                    "title": "Complex Dict",
                                    "type": "object",
                                },
                                "enum": {
                                    "allOf": [{"$ref": "#/$defs/StubEnum"}],
                                    "description": "I am an enum.",
                                },
                                "defaulted_str": {
                                    "default": "default",
                                    "description": "I have a string default value.",
                                    "title": "Defaulted Str",
                                    "type": "string",
                                },
                                "defaulted_float": {
                                    "default": 1.0,
                                    "description": "I have a float default value.",
                                    "title": "Defaulted Float",
                                    "type": "number",
                                },
                            },
                            "required": [
                                "x",
                                "y",
                                "union",
                                "pydantic_model",
                                "basic_dict",
                                "complex_dict",
                                "enum",
                            ],
                            "$defs": {
                                "StubEnum": {
                                    "description": "Stub enum docstring.",
                                    "enum": [1, 2],
                                    "title": "StubEnum",
                                    "type": "integer",
                                },
                                "StubState": {
                                    "description": "Stub model docstring.",
                                    "properties": {
                                        "defaulted_int": {
                                            "default": 1,
                                            "description": "A description of the int.",
                                            "title": "Defaulted Int",
                                            "type": "integer",
                                        },
                                        "required_str": {
                                            "description": "A description of the str.",
                                            "title": "Required Str",
                                            "type": "string",
                                        },
                                    },
                                    "required": ["required_str"],
                                    "title": "StubState",
                                    "type": "object",
                                },
                            },
                        },
                    },
                },
                id="many-edge-cases",
            ),
            pytest.param(
                add,
                {},
                {
                    "type": "function",
                    "info": {
                        "name": "add",
                        "description": "Add two numbers.",
                        "parameters": {
                            "properties": {
                                "a": {
                                    "description": "first number",
                                    "title": "A",
                                    "type": "integer",
                                },
                                "b": {
                                    "description": "second number",
                                    "title": "B",
                                    "type": "integer",
                                },
                            },
                            "required": ["a", "b"],
                            "type": "object",
                        },
                    },
                },
                id="with-args-and-returns",
            ),
            pytest.param(
                example_fxn,
                {},
                {
                    "type": "function",
                    "info": {
                        "name": "example_fxn",
                        "description": (
                            "A test function.\n\nThere may be non-summary content."
                        ),
                        "parameters": {
                            "properties": {
                                "x": {
                                    "description": "x",
                                    "title": "X",
                                    "type": "integer",
                                },
                                "y": {
                                    "description": "y",
                                    "title": "Y",
                                    "type": "string",
                                },
                                "z": {
                                    "description": "z",
                                    "title": "Z",
                                    "type": "number",
                                },
                            },
                            "required": ["x", "y", "z"],
                            "type": "object",
                        },
                    },
                },
                id="with-linefeed",
            ),
        ],
    )
    def test_from_function(
        self, fn: Callable, kwargs: dict[str, Any], expected: dict[str, Any]
    ) -> None:
        assert (
            Tool.from_function(fn, **kwargs).model_dump(exclude_none=True) == expected
        )

    @pytest.mark.parametrize(
        ("fn", "kwargs", "expected"),
        [
            (
                example_fxn,
                {},
                """NAME: example_fxn

SYNOPSIS:
    example_fxn(integer x, string y, number z)

DESCRIPTION:
    A test function.

    There may be non-summary content.

PARAMETERS:
    x (integer): x
    y (string): y
    z (number): z""",
            ),
            (
                intuitive_arg,
                {"allow_empty_param_descriptions": True},
                """NAME: intuitive_arg

SYNOPSIS:
    intuitive_arg(string x)

DESCRIPTION:
    Cast the input argument x to a float.

PARAMETERS:
    x (string): No description provided.""",
            ),
        ],
    )
    def test_describe_str(
        self, fn: Callable, kwargs: dict[str, Any], expected: str
    ) -> None:
        tool = Tool.from_function(fn, **kwargs)
        assert tool.info.describe_str().strip() == expected

    def test_describe(self, subtests: SubTests) -> None:
        """Test that describe_xyz functions for FunctionInfo are reasonable."""
        tool = Tool.from_function(example_fxn)

        with subtests.test("Test describe_xml is callable"):
            assert tool.info.describe_xml()

        with subtests.test("Test describe_json is callable"):
            assert tool.info.describe_json()

    def test_serialization_manual(self) -> None:
        # make one manually
        tool = Tool(
            tool_fn=add,
            info=FunctionInfo(
                name="add",
                description="Add two numbers.",
                parameters={
                    "properties": {
                        "a": {
                            "description": "first number",
                            "title": "A",
                            "type": "integer",
                        },
                        "b": {
                            "description": "second number",
                            "title": "B",
                            "type": "integer",
                        },
                    },
                    "required": ["a", "b"],
                    "type": "object",
                },
            ),
        )

        ref = json.loads(r"""{
        "type": "function",
        "function": {
            "name": "add",
            "description": "Add two numbers.",
            "parameters": {
                "type": "object",
                "properties": {
                    "a": {
                        "type": "integer",
                        "title": "A",
                        "description": "first number"
                    },
                    "b": {
                        "type": "integer",
                        "title": "B",
                        "description": "second number"
                    }
                },
                "required": [
                    "a",
                    "b"
                ]
            }
        }
    }""")
        # make sure it agrees with the reference
        my_dump = json.loads(tool.model_dump_json(exclude_none=True, by_alias=True))
        assert my_dump == ref

        # make one from a function
        tool_fxn = Tool.from_function(add)
        # make sure it serializes correctly
        assert tool_fxn.model_dump_json(
            exclude_none=True, by_alias=True
        ) == tool.model_dump_json(exclude_none=True, by_alias=True)

    @pytest.mark.asyncio
    async def test_arg_types(self) -> None:
        tool = Tool.from_function(example_fxn)

        assert tool.info.parameters.properties["x"]["type"] == "integer"
        assert tool.info.parameters.properties["y"]["type"] == "string"
        assert tool.info.parameters.properties["z"]["type"] == "number"

        calls = [
            ToolCall.from_name(tool.info.name, x=5, y="hi", z=4.2),
        ]
        for call in calls:
            # Call the function to make sure argument types
            # are passed correctly. Private because
            # it doesn't serialize
            tool._tool_fn(**call.function.arguments)

    @pytest.mark.asyncio
    async def test_tool_serialization(
        self, dummy_env: DummyEnv, subtests: SubTests
    ) -> None:
        def get_todo_list(n: int):
            """Get todo list for today.

            Args:
                n: number of items to return
            """
            return "\n".join(["Go for a walk", "Read a book", "Call a friend"][:n])

        tool = Tool.from_function(get_todo_list)

        with subtests.test("pickling"):
            # Check round-trip pickling doesn't break the original Tool
            orig_tool_fn_id = id(tool._tool_fn)
            pickle.loads(pickle.dumps(tool))  # noqa: S301
            assert id(tool._tool_fn) == orig_tool_fn_id

        with subtests.test("serialization then deserialization"):
            tool_copy = Tool(**tool.model_dump(by_alias=True))
            assert tool.type == tool_copy.type
            assert tool.info == tool_copy.info

        dummy_env.tools = [tool]

        with subtests.test("tool call from dump"):
            # Mimic the way an ToolCall might be invoked by an LLM API:
            # the arguments will be strings.
            action = ToolRequestMessage(**{  # noqa: PIE804
                "tool_calls": [
                    {
                        "id": "good_tool_call",
                        "function": {"name": "get_todo_list", "arguments": '{"n": 2}'},
                    },
                    {
                        "id": "bad_tool_call",
                        "function": {
                            "name": "get_todo_list",
                            "arguments": '({"n": 2})',  # NOTE: invalid JSON
                        },
                    },
                ]
            })

            assert action.tool_calls[0].function.arguments == {"n": 2}
            assert action.tool_calls[1].function.name == INVALID_TOOL_NAME

        with subtests.test("tool call from name"):
            tool_call = ToolCall.from_name("get_todo_list", n=2)
            action = ToolRequestMessage(tool_calls=[tool_call])
            new_messages = await dummy_env.exec_tool_calls(action)
            assert new_messages[0].content == "Go for a walk\nRead a book"

        with subtests.test("tool call from tool"):
            tool_call = ToolCall.from_tool(tool, n=2)
            action = ToolRequestMessage(tool_calls=[tool_call])
            new_messages = await dummy_env.exec_tool_calls(action)
            assert new_messages[0].content == "Go for a walk\nRead a book"

        with subtests.test("tool call from tool with no kwargs"):
            tool_call = ToolCall.from_tool(tool, 3)
            action = ToolRequestMessage(tool_calls=[tool_call])
            new_messages = await dummy_env.exec_tool_calls(action)
            assert (
                new_messages[0].content == "Go for a walk\nRead a book\nCall a friend"
            )

        def get_todo_list_no_args():
            """Get todo list for today."""
            return "Go for a walk"

        tool = Tool.from_function(get_todo_list_no_args)
        dummy_env.tools = [tool]

        with subtests.test("tool call from tool with no args and order mismatch"):
            tool_call = ToolCall.from_tool(tool)
            action = ToolRequestMessage(tool_calls=[tool_call])
            new_messages = await dummy_env.exec_tool_calls(action)
            assert new_messages[0].content == "Go for a walk"

            tool_call = ToolCall.from_tool(tool, 1, 10, 30441)
            action = ToolRequestMessage(tool_calls=[tool_call])
            new_messages = await dummy_env.exec_tool_calls(action)
            assert new_messages[0].content == "Go for a walk"


@pytest.mark.asyncio
async def test_argref_by_name() -> None:
    class MyState:
        def __init__(self):
            self.refs = {"foo": 1}

    # Check we can use argref_by_name to add 1 + 2 using a value in refs
    wrapped_add = argref_by_name()(add)
    s = MyState()
    result = wrapped_add("foo", 2, state=s)
    # Now s.refs has a new entry at the below `name`
    name = result.split()[0]
    assert s.refs[name] == 1 + 2

    # Check we can still use argref_by_name without refs
    result = wrapped_add(6, 2, state=s)
    assert s.refs[result.split()[0]] == 6 + 2

    # Check if we use a key name that doesn't exist, we blow up
    with pytest.raises(KeyError, match="not found in state"):
        wrapped_add("bar", 2, state=MyState())

    # Check if state doesn't have refs, we blow up
    with pytest.raises(AttributeError, match="must have a 'refs' attribute"):
        wrapped_add("foo", 2, state="not a state")

    # now try with async and decorator
    @argref_by_name()
    async def async_add(a: int, b: int) -> int:
        """Some docstring."""
        return a + b

    result = await async_add("foo", 2, state=s)
    assert s.refs[result.split()[0]] == 1 + 2
    result = await async_add(6, 2, state=s)
    assert s.refs[result.split()[0]] == 6 + 2

    # now try with lists
    s.refs["bar"] = 7
    result = await async_add("foo", "bar", state=s)
    assert s.refs[result.split()[0]] == 1 + 7

    # try the convenience of comma splitting on key
    result = await async_add("foo,bar", state=s)
    assert s.refs[result.split()[0]] == 1 + 7

    @argref_by_name()
    async def async_list(a: int, b: int) -> list[int]:
        """Some docstring."""
        return [a, b]

    result = await async_list("foo", 2, state=s)
    name1, name2 = (n.split()[0] for n in result.split("\n"))
    assert s.refs[name1] == 1
    assert s.refs[name2] == 2

    @argref_by_name(return_direct=True)
    async def async_list_direct(a: int, b: int) -> list[int]:
        """Some docstring."""
        return [a, b]

    assert await async_list_direct("foo", 2, state=s) == [1, 2]

    # call in context
    tool = Tool.from_function(argref_by_name()(add))

    tool_call = ToolCall.from_tool(tool, "foo", 2)
    action = ToolRequestMessage(tool_calls=[tool_call])
    my_env = DummyEnv()
    my_env.tools = [tool]
    new_messages = await my_env.exec_tool_calls(action, state=MyState())
    assert new_messages[0].content.endswith("3")

    # assert that we can describe the tool
    assert tool.info.describe_str()
    assert (
        "(set via a string key instead of the full object)" in tool.info.describe_str()
    )

    # now try state passing
    @argref_by_name(fxn_requires_state=True)
    async def want_state(a: int, state: MyState) -> int:  # noqa: ARG001
        """Some docstring.

        Args:
            a: first number
            state: the state object
        """
        return 2 * a

    tool = Tool.from_function(want_state)
    action = ToolRequestMessage(tool_calls=[ToolCall.from_tool(tool, "foo")])
    my_env = DummyEnv()
    my_env.tools = [tool]
    await my_env.exec_tool_calls(action, state=MyState())



import uuid
from functools import partial, update_wrapper

from aviary.utils import is_coroutine_callable


def make_pretty_id(prefix: str = "") -> str:
    """
    Get an ID that is made using an optional prefix followed by part of an uuid4.

    For example:
    - No prefix: "ff726cd1"
    - With prefix of "foo": "foo-ff726cd1"
    """
    uuid_frags: list[str] = str(uuid.uuid4()).split("-")
    if not prefix:
        return uuid_frags[0]
    return prefix + "-" + uuid_frags[0]


ARGREF_NOTE = "(set via a string key instead of the full object)"


def argref_wrapper(wrapper, wrapped):
    """Inject the ARGREF_NOTE into the Args."""
    # normal wraps
    wrapped_func = update_wrapper(wrapper, wrapped)
    # now adjust what we need
    for a in wrapped_func.__annotations__:
        if a in {"return", "state"}:
            continue
        wrapped_func.__annotations__[a] = str

    # now add note to docstring for all relevant Args
    ds = wrapped_func.__doc__
    if ds and "Args:" in ds:
        arg_doc = ds.split("Args:")[1].split("Returns:")[0].split("\n")
        for line in arg_doc:
            if line.strip():  # Filter whitespace
                ds = ds.replace(line, " ".join((line, ARGREF_NOTE)))
    wrapped_func.__doc__ = ds
    return wrapped_func


def argref_wraps(wrapped):
    """Enable decorator syntax with argref_wrapper."""
    return partial(argref_wrapper, wrapped=wrapped)


def argref_by_name(  # noqa: C901
    fxn_requires_state: bool = False, prefix: str = "", return_direct: bool = False
):
    """Decorator to allow args to be a string key into a refs dict instead of the full object.

    This can prevent LLM-powered tool selections from getting confused by full objects,
    instead it enables them to work using named references.

    Args:
        fxn_requires_state: Whether to pass the state object to the decorated function.
        prefix: A prefix to add to the generated reference ID.
        return_direct: Whether to return the result directly or update the state object.

    Example 1:
        >>> @argref_by_name()  # doctest: +SKIP
        >>> def my_func(foo: float): ...  # doctest: +SKIP

    Example 2:
        >>> def my_func(foo: float, bar: float) -> list[float]:
        ...     return [foo, bar]
        >>> wrapped_fxn = argref_by_name()(my_func)
        >>> # Equivalent to my_func(state.refs["foo"])
        >>> wrapped_fxn("foo", state=state)  # doctest: +SKIP

    Working with lists:
    - If you return a list, the decorator will create a new reference for each item in the list.
    - If you pass multiple args that are strings, the decorator will assume those are the keys.
    - If you need to pass a string, then use a keyword argument.

    Example 1:
        >>> @argref_by_name()  # doctest: +SKIP
        >>> def my_func(foo: float, bar: float) -> list[float]:  # doctest: +SKIP
        ...     return [foo, bar]  # doctest: +SKIP

    Example 2:
        >>> def my_func(foo: float, bar: float) -> list[float]:
        ...     return [foo, bar]
        >>> wrapped_fxn = argref_by_name()(my_func)
        >>> # Returns a multiline string with the new references
        >>> # Equivalent to my_func(state.refs["a"], state.refs["b"])
        >>> wrapped_fxn("a", "b", state=state)  # doctest: +SKIP
    """

    def decorator(func):  # noqa: C901
        def get_call_args(*args, **kwargs):
            if "state" not in kwargs:
                raise ValueError(
                    "argref_by_name decorated function must have a 'state' argument. "
                    f"Function signature: {func.__name__}({', '.join(func.__annotations__)}) "
                    f" received args: {args} kwargs: {kwargs}"
                )
            # pop the state argument
            state = kwargs["state"] if fxn_requires_state else kwargs.pop("state")

            # pop all string arguments
            other_args = []
            keyname_args = []
            deref_args = []
            largs = list(args)
            while largs and isinstance(largs[0], str):
                keyname_args.append(largs.pop(0))
            if largs:
                other_args.extend(largs)
            if not keyname_args and kwargs:
                name = next(iter(kwargs))
                keyname_args = [kwargs.pop(name)]

            # now convert the keynames to actual references
            for arg in keyname_args:
                try:
                    if arg in state.refs:
                        deref_args.append(state.refs[arg])
                    # sometimes it is not correctly converted to a tuple
                    # so as an attempt to be helpful...
                    elif all(a.strip() in state.refs for a in arg.split(",")):
                        deref_args.extend([
                            state.refs[a.strip()] for a in arg.split(",")
                        ])
                    else:
                        raise KeyError(
                            f"Key '{arg}' not found in state. Available keys: {list(state.refs.keys())}"
                        )
                except AttributeError as e:
                    raise AttributeError(
                        "The state object must have a 'refs' attribute to use argref_by_name decorator."
                    ) from e
            return deref_args, other_args, kwargs, state

        def update_state(state, result):
            if return_direct:
                return result
            # if it returns a list, rather than storing the list as a single reference
            # we store each item in the list as a separate reference
            if isinstance(result, list):
                msg = []
                for item in result:
                    new_name = make_pretty_id(prefix)
                    state.refs[new_name] = item
                    msg.append(f"{new_name} ({item.__class__.__name__}): {item!s}")
                return "\n".join(msg)
            new_name = make_pretty_id(prefix)
            state.refs[new_name] = result
            return f"{new_name} ({result.__class__.__name__}): {result!s}"

        @argref_wraps(func)
        def wrapper(*args, **kwargs):
            args, other_args, kwargs, state = get_call_args(*args, **kwargs)
            result = func(*args, *other_args, **kwargs)
            return update_state(state, result)

        @argref_wraps(func)
        async def awrapper(*args, **kwargs):
            args, other_args, kwargs, state = get_call_args(*args, **kwargs)
            result = await func(*args, *other_args, **kwargs)
            return update_state(state, result)

        wrapper.requires_state = True
        awrapper.requires_state = True
        if is_coroutine_callable(func):
            return awrapper
        return wrapper

    return decorator



import os
from pathlib import Path

import pytest

from aviary.db import EnvironmentDBBackend
from aviary.env import DummyEnv

TEST_OUTPUT_DIR = Path(__file__).parent / "test_outputs"

ENV_BACKENDS = (
    "sqlite+aiosqlite:///:memory:",
    f"sqlite+aiosqlite:///{TEST_OUTPUT_DIR}/envruns.sqlite3",
)


@pytest.fixture(name="dummy_env")
def fixture_dummy_env() -> DummyEnv:
    return DummyEnv()


@pytest.fixture(scope="session", autouse=True)
def _fixture_make_test_outputs_dir():
    TEST_OUTPUT_DIR.mkdir(exist_ok=True)


@pytest.fixture
def clean_db_backend(request):
    """Empty the DB session between test cases and remove the database file."""
    db_backend = request.param

    # Remove the database file if it exists
    if db_backend.startswith("sqlite") and not db_backend.endswith(":memory:"):
        file_path = db_backend.split("///")[-1]
        if os.path.exists(file_path):
            os.remove(file_path)

    yield db_backend

    # Finalizer code to run after the test
    EnvironmentDBBackend.Session = None
    EnvironmentDBBackend.uri = None



import base64
import contextlib
import inspect
import io
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, ClassVar

from pydantic import JsonValue
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker, create_async_engine
from sqlalchemy.ext.asyncio import AsyncSession as _AsyncSession
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm.decl_api import _TypeAnnotationMapType
from sqlalchemy.types import JSON

if TYPE_CHECKING:
    import numpy as np


def partial_format(value: str, **formats: dict[str, Any]) -> str:
    """Partially format a string given a variable amount of formats."""
    for template_key, template_value in formats.items():
        with contextlib.suppress(KeyError):
            value = value.format(**{template_key: template_value})
    return value


def encode_image_to_base64(img: "np.ndarray") -> str:
    """Encode an image to a base64 string, to be included as an image_url in a Message."""
    try:
        from PIL import Image
    except ImportError as e:
        raise ImportError(
            "Image processing requires the 'image' extra for 'Pillow'. Please:"
            " `pip install aviary[image]`."
        ) from e

    image = Image.fromarray(img)
    buffer = io.BytesIO()
    image.save(buffer, format="PNG")
    return (
        f'data:image/png;base64,{base64.b64encode(buffer.getvalue()).decode("utf-8")}'
    )


def is_coroutine_callable(obj) -> bool:
    """Get if the input object is awaitable."""
    if inspect.isfunction(obj) or inspect.ismethod(obj):
        return inspect.iscoroutinefunction(obj)
    if callable(obj):
        return inspect.iscoroutinefunction(obj.__call__)
    return False


class Base(DeclarativeBase):
    # Allowing dict to correspond with arbitrary JSON
    type_annotation_map: ClassVar[_TypeAnnotationMapType] = {JsonValue: JSON}


class DBBackend:
    engine: ClassVar[AsyncEngine | None] = None
    Session: ClassVar[async_sessionmaker[_AsyncSession] | None] = None
    uri: ClassVar[str | None] = None

    @classmethod
    async def populate_session(
        cls,
        uri: str,
        base: type[DeclarativeBase] = Base,
        engine_kwargs: dict[str, Any] | None = None,
        sessionmaker_kwargs: dict[str, Any] | None = None,
    ) -> None:
        cls.uri = uri
        cls.engine = engine = create_async_engine(cls.uri, **(engine_kwargs or {}))
        async with engine.begin() as conn:
            await conn.run_sync(base.metadata.create_all)
        cls.Session = async_sessionmaker(bind=engine, **(sessionmaker_kwargs or {}))

    @classmethod
    @asynccontextmanager
    async def begin_session(cls) -> AsyncIterator[_AsyncSession]:
        if cls.Session is None:
            raise RuntimeError(
                "No database connection established. Please call `await"
                " DBBackend.init_connection(...)` or manually populate `cls.Session`"
                " before attempting to persist or rehydrate contexts."
            )
        async with cls.Session.begin() as session:
            yield session



from __future__ import annotations

import logging
import uuid
from datetime import datetime
from typing import Any, ClassVar

from pydantic import BaseModel, ConfigDict, Field, JsonValue
from sqlalchemy import Column, DateTime, ForeignKey, select
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column
from sqlalchemy.orm.decl_api import _TypeAnnotationMapType
from sqlalchemy.sql import func
from sqlalchemy.types import JSON

from aviary.utils import DBBackend
from aviary.version import __version__

logger = logging.getLogger(__name__)


class EnvironmentBase(DeclarativeBase):
    # Allowing dict to correspond with arbitrary JSON
    type_annotation_map: ClassVar[_TypeAnnotationMapType] = {JsonValue: JSON}


class EnvironmentDBBackend(DBBackend):
    @classmethod
    async def get_environment_instances(
        cls, name: str | None = None, environment_id: uuid.UUID | None = None
    ) -> list[EnvironmentDBSchema]:
        """Given a name, pull associated environment instance data.

        Args:
            name: The name of the environment this instance is associated with
            environment_id: The UUID of a specific environment instances

        Returns:
           A list of environment UUIDs.
        """
        async with cls.begin_session() as session:
            if environment_id is not None:
                query = select(EnvironmentDB).where(EnvironmentDB.id == environment_id)
            elif name is not None:
                query = select(EnvironmentDB).where(EnvironmentDB.name == name)
            else:
                raise ValueError("Must provide either a name or environment_id")

            results = (await session.execute(query)).scalars().all()
            return [EnvironmentDBSchema.model_validate(result) for result in results]

    @classmethod
    async def get_environment_frames(
        cls, environment_id: uuid.UUID | None, frame_id: uuid.UUID | None
    ) -> list[FrameDBSchema]:
        """Given a environment_id and frame_id data, get environment frame data.

        Args:
            environment_id: The environment instance we are looking for
            frame_id: The specific frame we are looking for

        Returns:
           A list of FrameDB.
        """
        async with cls.begin_session() as session:
            if frame_id is not None:
                query = select(FrameDB).where(FrameDB.id == frame_id)
            elif environment_id is not None:
                query = select(FrameDB).where(FrameDB.environment_id == environment_id)
            else:
                raise ValueError("Must provide either a environment_id or frame_id")
            results = (await session.execute(query)).scalars().all()
            return [FrameDBSchema.model_validate(result) for result in results]

    @classmethod
    async def add_environment_frame(
        cls,
        environment_id: uuid.UUID,
        state: dict[str, Any],
        metadata: dict[str, Any] | None = None,
    ) -> uuid.UUID:
        frame_id = uuid.uuid4()
        async with cls.begin_session() as session:
            session.add(
                FrameDB(
                    id=frame_id,
                    environment_id=environment_id,
                    state=state,
                    supplemental_data=metadata,
                )
            )
        return frame_id

    @classmethod
    async def add_environment_instance(
        cls,
        name: str,
    ) -> uuid.UUID:
        environment_id = uuid.uuid4()
        async with cls.begin_session() as session:
            session.add(
                EnvironmentDB(id=environment_id, name=name, aviary_version=__version__)
            )
        return environment_id


class EnvironmentDB(EnvironmentBase):
    """Rows here correspond to a single environment trajectory, a single named set of frames."""

    __tablename__ = "environments"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True)
    created_at: Mapped[datetime] = mapped_column(default=func.now())
    name: Mapped[str] = mapped_column(index=True)
    aviary_version: Mapped[str] = mapped_column(default=__version__)


class FrameDB(EnvironmentBase):
    """Rows are a snapshot of an environment after each step.

    Note: corresponds to a frame in the renderer object

    """

    __tablename__ = "frames"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True)
    environment_id: Mapped[uuid.UUID] = mapped_column(
        ForeignKey(EnvironmentDB.id), index=True
    )
    state: Mapped[JsonValue]  # Just for human readability
    supplemental_data: Mapped[JsonValue]
    created_at = Column(DateTime(timezone=True), server_default=func.now())


class EnvironmentDBSchema(BaseModel):
    """Sister model for EnvironmentDB."""

    id: uuid.UUID
    created_at: datetime = Field(default_factory=datetime.now)
    name: str
    aviary_version: str = __version__

    model_config = ConfigDict(from_attributes=True)


class FrameDBSchema(BaseModel):
    """Sister model for FrameDB."""

    id: uuid.UUID
    environment_id: uuid.UUID
    state: JsonValue
    supplemental_data: JsonValue
    created_at: datetime = Field(default_factory=datetime.now)

    model_config = ConfigDict(from_attributes=True)



import asyncio
import json
import os
import pathlib
import tempfile
from typing import ClassVar
from unittest.mock import patch

import pytest
from httpx import AsyncClient
from pydantic import BaseModel

from aviary.api import EnvDBClient, make_environment_db_server
from aviary.env import DummyEnv, DummyEnvState, Environment, Frame, TaskDataset
from aviary.message import Message
from aviary.render import Renderer
from aviary.tools import (
    Tool,
    ToolCall,
    ToolRequestMessage,
    ToolResponseMessage,
    ToolsAdapter,
)
from tests import CILLMModelNames
from tests.conftest import ENV_BACKENDS


class TestDummyEnv:
    @pytest.mark.asyncio
    async def test_dummyenv(self, dummy_env: DummyEnv) -> None:
        async def my_policy(obs: list[Message]) -> ToolRequestMessage:  # noqa: ARG001
            # For testing purposes, we hardcoded the policy
            return ToolRequestMessage(
                tool_calls=[
                    ToolCall.from_name("print_story", story="Once upon a time done")
                ],
            )

        obs, _ = await dummy_env.reset()
        assert isinstance(obs, list)
        assert len(obs) == 1

        action = await my_policy(obs)
        _, reward, done, _ = await dummy_env.step(action)
        assert reward > 0
        assert done

    @pytest.mark.asyncio
    async def test_tool_signatures(self, dummy_env: DummyEnv) -> None:
        _, tools = await dummy_env.reset()
        assert ToolsAdapter.dump_python(tools, exclude_none=True) == [
            {
                "type": "function",
                "info": {
                    "name": "print_story",
                    "description": "Print a story.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "story": {
                                "type": "string",
                                "title": "Story",
                                "description": "Story to print.",
                            }
                        },
                        "required": ["story"],
                    },
                },
            },
            {
                "info": {
                    "description": "Cast the input argument x to a float.",
                    "name": "cast_float",
                    "parameters": {
                        "properties": {"x": {"type": "string", "title": "X"}},
                        "required": ["x"],
                        "type": "object",
                    },
                },
                "type": "function",
            },
            {
                "info": {
                    "description": "Cast the input argument x to an integer.",
                    "name": "cast_int",
                    "parameters": {
                        "properties": {"x": {"type": "number", "title": "X"}},
                        "required": ["x"],
                        "type": "object",
                    },
                },
                "type": "function",
            },
        ]

    def test_loading_from_name(self):
        env: DummyEnv = Environment.from_name("dummy")
        assert isinstance(env, DummyEnv)

        dataset = TaskDataset.from_name("dummy")
        batch = next(iter(dataset.iter_batches(1)))
        assert len(batch) == 1
        assert isinstance(batch[0], DummyEnv)

    @pytest.mark.parametrize(
        "model_name", ["gpt-3.5-turbo", CILLMModelNames.ANTHROPIC.value]
    )
    @pytest.mark.asyncio
    async def test_tool_calling(self, dummy_env: DummyEnv, model_name: str) -> None:
        def get_todo_list(n: int) -> str:
            """Get todo list for today.

            Args:
                n: number of items to return
            """
            return "\n".join(["Go for a walk", "Read a book", "Call a friend"][:n])

        tool = Tool.from_function(get_todo_list)
        dummy_env.tools = [tool]
        tool_request_message = ToolRequestMessage(
            tool_calls=[ToolCall.from_name("get_todo_list", n=3)]
        )
        new_messages = await dummy_env.exec_tool_calls(tool_request_message)
        (new_message,) = new_messages
        assert new_message.content == "Go for a walk\nRead a book\nCall a friend"
        assert new_message.tool_call_id == tool_request_message.tool_calls[0].id

        def get_todo_list_no_args() -> str:
            """Get todo list for today."""
            return "\n".join(["Go for a walk", "Read a book", "Call a friend"])

        tool = Tool.from_function(get_todo_list_no_args)
        dummy_env.tools = [tool]
        tool_request_message = ToolRequestMessage(
            tool_calls=[ToolCall.from_name("get_todo_list_no_args")]
        )
        new_messages = await dummy_env.exec_tool_calls(tool_request_message)
        (new_message,) = new_messages
        assert new_message.content == "Go for a walk\nRead a book\nCall a friend"
        assert new_message.tool_call_id == tool_request_message.tool_calls[0].id

        # ok now try with multiple functions

        def get_calendar() -> str:
            """Get text version of calendar for today."""
            return "9:00am Wake-up\n10:00pm Go to bed\n"

        tool2 = Tool.from_function(get_calendar)
        dummy_env.tools = [tool, tool2]
        tool_request_message = ToolRequestMessage(
            tool_calls=[
                ToolCall.from_name("get_todo_list_no_args"),
                ToolCall.from_name("get_calendar"),
            ],
        )
        new_messages = await dummy_env.exec_tool_calls(tool_request_message)
        if model_name.startswith("claude"):
            # Anthropic not always so smart
            assert 1 <= len(new_messages) <= 2
        else:
            assert len(new_messages) == 2


@pytest.mark.asyncio
async def test_multiple_calls(dummy_env: DummyEnv) -> None:
    obs, tools = await dummy_env.reset()
    calls = [
        ToolCall.from_name(tools[0].info.name, story="Hello, how are you?"),
        ToolCall.from_name(tools[0].info.name, story="Hello, how are you?"),
        ToolCall.from_name(tools[0].info.name, story="Hello, how are you?"),
    ]
    action = ToolRequestMessage(tool_calls=calls)
    obs, reward, done, truncated = await dummy_env.step(action)
    assert reward > 0
    assert done


class TestRendering:
    class SomeState(BaseModel):
        field: int

    @pytest.mark.parametrize(
        ("state", "serialized"),
        [
            (5, 5),
            (5.6, 5.6),
            ("hi", "hi"),
            (True, True),
            (["hi"], ["hi"]),
            ({"hi": 5}, {"hi": 5}),
            (SomeState(field=5), {"field": 5}),
        ],
    )
    def test_serialization(self, state, serialized) -> None:
        assert Frame(state=state).model_dump()["state"] == serialized

    def test_frame_mutability(self) -> None:
        # make a nested list - so shallow copy won't catch it
        mutable_state = [["foo"]]
        non_deep_copy = Frame(state=mutable_state, deepcopy=False)
        mutable_state[0].append("bar")
        assert non_deep_copy.model_dump()["state"] == [["foo", "bar"]]

        mutable_state = [["foo"]]
        deep_copy = Frame(state=mutable_state)
        mutable_state[0].append("bar")
        assert deep_copy.model_dump()["state"] == [["foo"]]

    def test_rendering(self, dummy_env: DummyEnv) -> None:
        # Reset to add state
        asyncio.run(dummy_env.reset())

        renderer = Renderer(name="Name", prefix="test")
        renderer.append(dummy_env.export_frame())
        with tempfile.TemporaryDirectory() as tmpdir:
            build_dir = pathlib.Path(tmpdir)
            renderer.build(build_dir)
            file_paths = list(build_dir.glob("*.json"))
            assert len(file_paths) == 2, "Expected manifest and one object"
            frame_path = file_paths[
                file_paths[0].name.removeprefix("test_").startswith("info")
            ]
            with frame_path.open() as f:
                rehydrated = json.load(f)
        assert rehydrated["state"]["messages"] == [
            "Write a 5 word story via print_story"
        ]


class ParallelizedDummyEnv(DummyEnv):
    def __init__(self, right_hand_broken: bool = False):
        super().__init__()
        self.right_hand_broken = right_hand_broken

    RIGHT_HAND_BROKEN_MESSAGE: ClassVar[str] = "Right hand is broken."

    async def reset(self) -> tuple[list[Message], list[Tool]]:
        def move_right_hand(
            distance: int,  # noqa: ARG001
            state: DummyEnvState,
        ) -> None:
            """
            Move your right hand forward or backward.

            Args:
                distance: Integer distance to move (mm), where forward is positive.
                state: Current state.
            """
            if self.right_hand_broken:  # Use this to test tool errors
                raise RuntimeError(self.RIGHT_HAND_BROKEN_MESSAGE)
            state.reward += 1

        def move_left_hand(
            distance: int,  # noqa: ARG001
            state: DummyEnvState,
        ) -> None:
            """
            Move your left hand forward or backward.

            Args:
                distance: Integer distance to move (mm), where forward is positive.
                state: Current state.
            """
            state.reward += 1

        def smile_and_wave(state: DummyEnvState) -> None:
            """
            Smile and wave.

            Args:
                state: Current state.
            """
            state.reward = 10
            state.done = True

        self.tools = [
            Tool.from_function(move_left_hand),
            Tool.from_function(move_right_hand),
            Tool.from_function(smile_and_wave),
        ]
        self.state = type(self).State(
            messages=[
                Message(
                    role="user",
                    content=(
                        "You are the president of the United States of America."
                        " Please move both hands at the same time, and then smile"
                        " and wave."
                    ),
                )
            ]
        )
        return self.state.messages, self.tools


class TestParallelism:
    @pytest.mark.parametrize(
        "model_name", [CILLMModelNames.ANTHROPIC.value, "gpt-4-turbo"]
    )
    @pytest.mark.asyncio
    async def test_exec_tool_calls_handling(self, model_name: str) -> None:
        env = ParallelizedDummyEnv(right_hand_broken=True)
        obs, tools = await env.reset()
        right_hand_tool = tools[1]

        # 1. Let's DIY create a ToolRequestMessage for test determinism
        request_msg = ToolRequestMessage(
            tool_calls=[ToolCall.from_tool(right_hand_tool, distance=5)]
        )

        # 2. Okay, our hand was broken, let's handle it DIY-style
        try:
            obs, *_ = await env.step(action=request_msg)
        except RuntimeError as exc:
            obs = [
                Message(
                    content=f"Failed to execute tools with message:\n{exc}", role="tool"
                )
            ]
        else:
            raise AssertionError("Should have blown up per the test logic.")

        # 2. Now that we have confirmed that, let's make sure exec_tool_calls
        #    can automate this for us
        obs = await env.exec_tool_calls(
            message=request_msg, state=env.state, handle_tool_exc=True
        )
        (failure_tool_response,) = obs
        assert isinstance(failure_tool_response, ToolResponseMessage)
        assert env.RIGHT_HAND_BROKEN_MESSAGE in failure_tool_response.content


class TestEnvDBServerClient:
    @pytest.mark.parametrize("clean_db_backend", ENV_BACKENDS, indirect=True)
    @pytest.mark.asyncio
    async def test_env_write_and_read(self, clean_db_backend: str) -> None:
        base_server_url = "http://testserver"
        test_env_name = "test_env"
        with patch.dict(
            os.environ, {"AUTH_TOKEN": "stub", "ENV_DB_URI": clean_db_backend}
        ):
            async with AsyncClient(
                app=make_environment_db_server(),
                base_url=base_server_url,
            ) as async_client:
                response = await async_client.post(
                    "/environment_instance",
                    headers={"Authorization": "Bearer stub"},
                    params={"env_name": test_env_name},
                )
                response.raise_for_status()

            custom_client = AsyncClient(
                app=make_environment_db_server(),
                base_url=base_server_url,
            )

            env_db_client = EnvDBClient(
                server_url="",
                request_headers={"Authorization": "Bearer stub"},
            )

            with patch("httpx.AsyncClient.get", custom_client.get):
                all_results = await env_db_client.get_environment_instances(
                    name=test_env_name
                )
                assert response.json() == all_results[0].id

    @pytest.mark.parametrize("clean_db_backend", ENV_BACKENDS, indirect=True)
    @pytest.mark.asyncio
    async def test_env_frame_write_and_read(self, clean_db_backend: str) -> None:
        base_server_url = "http://testserver"
        test_env_name = "test_env"
        with patch.dict(
            os.environ, {"AUTH_TOKEN": "stub", "ENV_DB_URI": clean_db_backend}
        ):
            custom_client = AsyncClient(
                app=make_environment_db_server(),
                base_url=base_server_url,
            )

            env_db_client = EnvDBClient(
                server_url="",
                request_headers={"Authorization": "Bearer stub"},
                # custom_client_func=custom_client_func,
            )
            with (
                patch("httpx.AsyncClient.post", custom_client.post),
                patch("httpx.AsyncClient.get", custom_client.get),
            ):
                environment_id = await env_db_client.write_environment_instance(
                    name=test_env_name
                )
                environment_ids = await env_db_client.get_environment_instances(
                    name=test_env_name
                )

                assert (
                    len(environment_ids) == 1
                ), "There should only be 1 instance present"

                frame = Frame(state={"messages": ["Hello, world!"]})
                frame_id_1 = await env_db_client.write_environment_frame(
                    environment_id=environment_id, frame=frame
                )
                frame_id_2 = await env_db_client.write_environment_frame(
                    environment_id=environment_id, frame=frame
                )
                frames = await env_db_client.get_environment_frames(
                    environment_id=environment_id
                )

                assert len(frames) == 2, "There should be 2 frames present"
                assert frames[0].id == str(
                    frame_id_1
                ), "Frame ID was not assigned as expected"
                assert frames[1].id == str(
                    frame_id_2
                ), "Frame ID was not assigned as expected"



import os
import secrets
import uuid
from typing import TYPE_CHECKING, Annotated, Any

import httpx

from aviary.db import (
    EnvironmentBase,
    EnvironmentDB,
    EnvironmentDBBackend,
    EnvironmentDBSchema,
    FrameDB,
    FrameDBSchema,
)
from aviary.render import Frame

if TYPE_CHECKING:
    from fastapi import FastAPI


class EnvDBClient:
    """Interact with an Environment DB via REST API."""

    def __init__(
        self,
        server_url: str,
        request_headers: httpx._types.HeaderTypes | None = None,
        request_timeout: float | None = None,
    ):
        self._request_url = server_url
        self._request_headers = request_headers
        self._request_timeout = request_timeout

    async def write_environment_instance(self, name: str) -> uuid.UUID:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                f"{self._request_url}/environment_instance",
                params={
                    "env_name": name,
                },
                headers=self._request_headers,
                timeout=self._request_timeout,
            )
            response.raise_for_status()
            return uuid.UUID(response.json())

    async def write_environment_frame(
        self, environment_id: str | uuid.UUID, frame: Frame
    ) -> uuid.UUID:
        async with httpx.AsyncClient() as client:
            frame_data = frame.model_dump()
            response = await client.post(
                f"{self._request_url}/environment_frame",
                params={"environment_id": str(environment_id)},
                json={
                    "state": frame_data.get("state"),
                    "metadata": frame_data.get("info"),
                },
                headers=self._request_headers,
                timeout=self._request_timeout,
            )
            response.raise_for_status()
            return uuid.UUID(response.json())

    async def get_environment_instances(
        self,
        name: str | None = None,
        environment_id: str | uuid.UUID | None = None,
    ) -> list[EnvironmentDB]:
        async with httpx.AsyncClient() as client:
            params = (
                {"env_name": name} if name else {"environment_id": str(environment_id)}
            )
            response = await client.get(
                f"{self._request_url}/environment_instance",
                params=params,
                headers=self._request_headers,
                timeout=self._request_timeout,
            )
            response.raise_for_status()
            return [EnvironmentDB(**obj) for obj in response.json()]

    async def get_environment_frames(
        self,
        environment_id: str | uuid.UUID | None = None,
        frame_id: str | uuid.UUID | None = None,
    ) -> list[FrameDB]:
        async with httpx.AsyncClient() as client:
            params = (
                {"environment_id": str(environment_id)}
                if environment_id
                else {"frame_id": str(frame_id)}
            )
            response = await client.get(
                f"{self._request_url}/environment_frame",
                params=params,
                headers=self._request_headers,
                timeout=self._request_timeout,
            )
            response.raise_for_status()
            return [FrameDB(**obj) for obj in response.json()]


def make_environment_db_server(render_docs: bool = False) -> "FastAPI":  # noqa: C901
    """Make a FastAPI app, an interface for the Environment DB."""
    try:
        from fastapi import Depends, FastAPI, HTTPException, status
        from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
    except ModuleNotFoundError as exc:
        raise ImportError(
            "Environment DB server requires the 'server' extra for 'fastapi'. Please:"
            " `pip install aviary[server]`."
        ) from exc

    backend = EnvironmentDBBackend()

    async def ensure_session():
        if not backend.Session:
            await backend.populate_session(
                base=EnvironmentBase, uri=os.environ["ENV_DB_URI"]
            )
        yield

    asgi_app = FastAPI(
        title="Aviary Environment DB API",
        description="CRUD operations for the Aviary Environment DB",
        # Only render Swagger docs if local since we don't have a login here
        docs_url="/docs" if render_docs else None,
        redoc_url="/redoc" if render_docs else None,
    )
    auth_scheme = HTTPBearer()

    async def validate_token(
        token: Annotated[HTTPAuthorizationCredentials, Depends(auth_scheme)],
    ) -> HTTPAuthorizationCredentials:
        # NOTE: don't use os.environ.get() to avoid possible empty string matches, and
        # to have clearer server failures if the AUTH_TOKEN env var isn't present
        if not secrets.compare_digest(token.credentials, os.environ["AUTH_TOKEN"]):
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Incorrect bearer token",
                headers={"WWW-Authenticate": "Bearer"},
            )
        return token

    @asgi_app.post("/environment_instance")
    async def add_environment_instance(
        env_name: str,
        _: Annotated[HTTPAuthorizationCredentials, Depends(validate_token)],
        _db: Annotated[None, Depends(ensure_session)],
    ) -> uuid.UUID:
        return await backend.add_environment_instance(name=env_name)

    @asgi_app.post("/environment_frame")
    async def add_environment_frame(
        environment_id: uuid.UUID,
        state: dict[str, Any],
        _: Annotated[HTTPAuthorizationCredentials, Depends(validate_token)],
        _db: Annotated[None, Depends(ensure_session)],
        metadata: dict[str, Any] | None = None,
    ) -> uuid.UUID:
        return await backend.add_environment_frame(
            environment_id=environment_id, state=state, metadata=metadata
        )

    @asgi_app.get("/environment_instance")
    async def get_environment_instances(
        _: Annotated[HTTPAuthorizationCredentials, Depends(validate_token)],
        _db: Annotated[None, Depends(ensure_session)],
        env_name: str | None = None,
        environment_id: uuid.UUID | None = None,
    ) -> list[EnvironmentDBSchema]:
        if not (env_name or environment_id):
            raise HTTPException(400, "Must provide either a name or environment_id")
        return await backend.get_environment_instances(
            name=env_name, environment_id=environment_id
        )

    @asgi_app.get("/environment_frame")
    async def get_environment_frames(
        _: Annotated[HTTPAuthorizationCredentials, Depends(validate_token)],
        _db: Annotated[None, Depends(ensure_session)],
        environment_id: uuid.UUID | None = None,
        frame_id: uuid.UUID | None = None,
    ) -> list[FrameDBSchema]:
        if not (frame_id or environment_id):
            raise HTTPException(400, "Must provide either a frame_id or environment_id")
        return await backend.get_environment_frames(
            environment_id=environment_id, frame_id=frame_id
        )

    return asgi_app
